@@ -1,21 +1,22 @@
-/* A Bison parser, made by GNU Bison 2.7.  */
+/* A Bison parser, made by GNU Bison 3.8.2.  */
 
 /* Bison interface for Yacc-like parsers in C
-   
-      Copyright (C) 1984, 1989-1990, 2000-2012 Free Software Foundation, Inc.
-   
+
+   Copyright (C) 1984, 1989-1990, 2000-2015, 2018-2021 Free Software Foundation,
+   Inc.
+
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* As a special exception, you may create a larger work that contains
    part or all of the Bison parser skeleton and distribute that work
@@ -26,13 +27,17 @@
    special exception, which will cause the skeleton and the resulting
    Bison output files to be licensed under the GNU General Public
    License without this special exception.
-   
+
    This special exception was added by the Free Software Foundation in
    version 2.2 of Bison.  */
 
+/* DO NOT RELY ON FEATURES THAT ARE NOT DOCUMENTED in the manual,
+   especially those whose name start with YY_ or yy_.  They are
+   private implementation details that can be changed or removed.  */
+
 #ifndef YY_YY_CFGRAM_H_INCLUDED
 # define YY_YY_CFGRAM_H_INCLUDED
-/* Enabling traces.  */
+/* Debug traces.  */
 #ifndef YYDEBUG
 # define YYDEBUG 1
 #endif
@@ -40,55 +45,64 @@
 extern int yydebug;
 #endif
 
-/* Tokens.  */
+/* Token kinds.  */
 #ifndef YYTOKENTYPE
 # define YYTOKENTYPE
-   /* Put the tokens into the symbol table, so that GDB and other debuggers
-      know about them.  */
-   enum yytokentype {
-     STRING = 258,
-     IDENT = 259,
-     NUMBER = 260,
-     RUSH = 261,
-     T_VERSION = 262,
-     RULE = 263,
-     GLOBAL = 264,
-     EOL = 265,
-     SET = 266,
-     INSERT = 267,
-     REMOPT = 268,
-     MAP = 269,
-     UNSET = 270,
-     MATCH = 271,
-     FALLTHROUGH = 272,
-     INCLUDE = 273,
-     LIMITS = 274,
-     CLRENV = 275,
-     SETENV = 276,
-     UNSETENV = 277,
-     KEEPENV = 278,
-     EVALENV = 279,
-     DELETE = 280,
-     EXIT = 281,
-     ATTRIB = 282,
-     GLATTRIB = 283,
-     BOGUS = 284,
-     OR = 285,
-     AND = 286,
-     NOT = 287,
-     EQ = 288,
-     NE = 289,
-     LT = 290,
-     LE = 291,
-     GT = 292,
-     GE = 293,
-     XF = 294,
-     NM = 295,
-     IN = 296,
-     GROUP = 297
-   };
+  enum yytokentype
+  {
+    YYEMPTY = -2,
+    YYEOF = 0,                     /* "end of file"  */
+    YYerror = 256,                 /* error  */
+    YYUNDEF = 257,                 /* "invalid token"  */
+    STRING = 258,                  /* "string"  */
+    IDENT = 259,                   /* "identifier"  */
+    NUMBER = 260,                  /* "number"  */
+    RUSH = 261,                    /* "rush"  */
+    T_VERSION = 262,               /* T_VERSION  */
+    RULE = 263,                    /* "rule"  */
+    GLOBAL = 264,                  /* "global"  */
+    EOL = 265,                     /* "end of line"  */
+    SET = 266,                     /* "set"  */
+    INSERT = 267,                  /* "insert"  */
+    REMOPT = 268,                  /* "remopt"  */
+    MAP = 269,                     /* "map"  */
+    UNSET = 270,                   /* "unset"  */
+    MATCH = 271,                   /* "match"  */
+    FALLTHROUGH = 272,             /* "fallthrough"  */
+    INCLUDE = 273,                 /* "include"  */
+    LIMITS = 274,                  /* "limits"  */
+    CLRENV = 275,                  /* "clrenv"  */
+    SETENV = 276,                  /* "setenv"  */
+    UNSETENV = 277,                /* "unsetenv"  */
+    KEEPENV = 278,                 /* "keepenv"  */
+    EVALENV = 279,                 /* "evalenv"  */
+    DELETE = 280,                  /* "delete"  */
+    EXIT = 281,                    /* "exit"  */
+    ATTRIB = 282,                  /* "rule attribute"  */
+    GLATTRIB = 283,                /* "global attribute"  */
+    BOGUS = 284,                   /* "erroneous token"  */
+    OR = 285,                      /* "||"  */
+    AND = 286,                     /* "&&"  */
+    NOT = 287,                     /* "!"  */
+    EQ = 288,                      /* "=="  */
+    NE = 289,                      /* "!="  */
+    LT = 290,                      /* "<"  */
+    LE = 291,                      /* "<="  */
+    GT = 292,                      /* ">"  */
+    GE = 293,                      /* ">="  */
+    XF = 294,                      /* "=~"  */
+    NM = 295,                      /* "!~"  */
+    IN = 296,                      /* "in"  */
+    GROUP = 297,                   /* "group"  */
+    TEST = 298                     /* "test"  */
+  };
+  typedef enum yytokentype yytoken_kind_t;
 #endif
-/* Tokens.  */
+/* Token kinds.  */
+#define YYEMPTY -2
+#define YYEOF 0
+#define YYerror 256
+#define YYUNDEF 257
 #define STRING 258
 #define IDENT 259
 #define NUMBER 260
@@ -129,14 +143,13 @@ extern int yydebug;
 #define NM 295
 #define IN 296
 #define GROUP 297
+#define TEST 298
 
-
-
+/* Value type.  */
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
-typedef union YYSTYPE
+union YYSTYPE
 {
-/* Line 2058 of yacc.c  */
-#line 37 "cfgram.y"
+#line 39 "cfgram.y"
 
 	char *str;
 	struct cfnumber num;
@@ -167,43 +180,35 @@ typedef union YYSTYPE
 		struct argval *tail;
 	} arglist;
 	struct { unsigned major, minor; } version;
+	int fstest;
 
+#line 186 "cfgram.h"
 
-/* Line 2058 of yacc.c  */
-#line 174 "cfgram.h"
-} YYSTYPE;
+};
+typedef union YYSTYPE YYSTYPE;
 # define YYSTYPE_IS_TRIVIAL 1
-# define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
 #endif
 
+/* Location type.  */
 #if ! defined YYLTYPE && ! defined YYLTYPE_IS_DECLARED
-typedef struct YYLTYPE
+typedef struct YYLTYPE YYLTYPE;
+struct YYLTYPE
 {
   int first_line;
   int first_column;
   int last_line;
   int last_column;
-} YYLTYPE;
-# define yyltype YYLTYPE /* obsolescent; will be withdrawn */
+};
 # define YYLTYPE_IS_DECLARED 1
 # define YYLTYPE_IS_TRIVIAL 1
 #endif
 
+
 extern YYSTYPE yylval;
 extern YYLTYPE yylloc;
-#ifdef YYPARSE_PARAM
-#if defined __STDC__ || defined __cplusplus
-int yyparse (void *YYPARSE_PARAM);
-#else
-int yyparse ();
-#endif
-#else /* ! YYPARSE_PARAM */
-#if defined __STDC__ || defined __cplusplus
+
 int yyparse (void);
-#else
-int yyparse ();
-#endif
-#endif /* ! YYPARSE_PARAM */
+
 
 #endif /* !YY_YY_CFGRAM_H_INCLUDED  */
