@@ -1,5 +1,5 @@
-/* This file is part of GNU Rush.                  
-   Copyright (C) 2008-2019 Sergey Poznyakoff
+/* This file is part of GNU Rush.
+   Copyright (C) 2008-2025 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,6 +16,7 @@
 
 #include <rush.h>
 #include <cf.h>
+#include <parseopt.h>
 
 extern char **environ;
 
@@ -39,27 +40,27 @@ struct error_msg {
 };
 
 struct error_msg error_msg[] = {
-        /* usage_error */
-	{ N_("You are not permitted to execute this command.\n"             
+	/* usage_error */
+	{ N_("You are not permitted to execute this command.\n"
 	     "Contact the systems administrator for further assistance.\n"), },
-        
-        /* nologin_error */
-        { N_("You do not have interactive login access to this machine.\n" 
+
+	/* nologin_error */
+	{ N_("You do not have interactive login access to this machine.\n"
 	     "Contact the systems administrator for further assistance.\n") },
-        
-        /* config_error */
-        { N_("Local configuration error occurred.\n" 
+
+	/* config_error */
+	{ N_("Local configuration error occurred.\n"
 	     "Contact the systems administrator for further assistance.\n") },
-        
-        /* system_error */
-        { N_("A system error occurred while attempting to execute command.\n" 
+
+	/* system_error */
+	{ N_("A system error occurred while attempting to execute command.\n"
 	     "Contact the systems administrator for further assistance.\n") }
 };
 
 void
 set_error_msg(enum error_type type, char *text)
 {
-	error_msg[type].text = text;
+	error_msg[type].text = xstrdup(text);
 	error_msg[type].custom = 1;
 }
 
@@ -104,7 +105,7 @@ new_error(int fd, char const *text, int unescape)
 	int add_nl = len > 0 && text[len-1] != '\n';
 	int c;
 	char *p;
-	
+
 	err = xmalloc(sizeof(*err) + strlen(text) + (add_nl ? 1 : 0) + 1);
 	err->fd = fd;
 	err->idx = -1;
@@ -112,31 +113,31 @@ new_error(int fd, char const *text, int unescape)
 	while ((c = *text++) != 0) {
 		if (unescape && c == '\\' && *text) {
 			int c1 = wordsplit_c_unquote_char(*text);
-			if (c1) 
+			if (c1)
 				c = c1;
 			else
 				c = *text;
 			text++;
-		}	
+		}
 		*p++ = c;
 	}
 	if (add_nl)
 		*p++ = '\n';
 	*p = 0;
-	
+
 	return err;
 }
 
 char const *
 rush_error_msg(struct rush_error const *err, struct rush_i18n const *i18n)
-{	
+{
 	const char *msg;
 	if (err->idx >= 0) {
 		msg = error_msg[err->idx].text;
 		if (error_msg[err->idx].custom) {
 			/* If it is a customized version, translate it via
 			   user-supplied i18n */
-			if (i18n) 
+			if (i18n)
 				msg = user_gettext(i18n->locale,
 						   i18n->text_domain,
 						   i18n->localedir,
@@ -151,7 +152,7 @@ rush_error_msg(struct rush_error const *err, struct rush_i18n const *i18n)
 void
 send_msg(const char *msg, size_t len)
 {
-        if (write(STDERR_FILENO, msg, len) < 0) {
+	if (write(STDERR_FILENO, msg, len) < 0) {
 		logmsg(LOG_ERR,
 		       _("failed to write message to stderr: %s"),
 		       strerror(errno));
@@ -171,11 +172,11 @@ vlogmsg(int prio, const char *fmt, va_list ap)
 		case LOG_DEBUG:
 			fprintf(stderr, _("Debug: "));
 			break;
-			
-		case LOG_INFO:      
+
+		case LOG_INFO:
 			fprintf(stderr, _("Info: "));
 			break;
-			
+
 		case LOG_NOTICE:
 			fprintf(stderr, _("Notice: "));
 			break;
@@ -183,10 +184,10 @@ vlogmsg(int prio, const char *fmt, va_list ap)
 		case LOG_WARNING:
 			fprintf(stderr, _("Warning: "));
 			break;
-			
-		case LOG_ERR:       
-		case LOG_CRIT:      
-		case LOG_ALERT:     
+
+		case LOG_ERR:
+		case LOG_CRIT:
+		case LOG_ALERT:
 		case LOG_EMERG:
 			fprintf(stderr, _("Error: "));
 		}
@@ -219,7 +220,7 @@ die(enum error_type type, struct rush_i18n *i18n, const char *fmt, ...)
 		if (error_msg[type].custom) {
 			/* If it is a customized version, translate it via
 			   user-supplied i18n */
-			if (i18n) 
+			if (i18n)
 				msg = user_gettext(i18n->locale,
 						   i18n->text_domain,
 						   i18n->localedir,
@@ -229,14 +230,14 @@ die(enum error_type type, struct rush_i18n *i18n, const char *fmt, ...)
 		send_msg(msg, strlen(msg));
 		sleep(sleep_time);
 	}
-        exit(1);
+	exit(1);
 }
 
 void
 die_usage(struct cfloc const *loc, char const *fmt, ...)
 {
-        va_list ap;
-        va_start(ap, fmt);
+	va_list ap;
+	va_start(ap, fmt);
 	vcferror(loc, fmt, ap);
 	va_end(ap);
 	die(usage_error, NULL, NULL);
@@ -245,7 +246,7 @@ die_usage(struct cfloc const *loc, char const *fmt, ...)
 void
 xalloc_die()
 {
-        die(system_error, NULL, _("Not enough memory"));
+	die(system_error, NULL, _("Not enough memory"));
 }
 
 static rush_bool_t
@@ -308,7 +309,7 @@ eval_cmps(struct test_node *node,
 {
 	char *str = rush_expand_string(node->v.cmp.larg, req);
 	rush_bool_t res = rush_false;
-	
+
 	switch (node->v.cmp.op) {
 	case cmp_eq:
 		res = strcmp(str, node->v.cmp.rarg.str) == 0;
@@ -344,23 +345,22 @@ eval_in(struct test_node *node,
 	free(str);
 	return res;
 }
-	
+
 static rush_bool_t
-groupmember(char const *gname, struct passwd const *pw)
+groupmember(struct group *grp, struct passwd const *pw)
 {
-        struct group *grp = getgrnam(gname);
-        if (grp) {
+	if (grp) {
 		char **p;
-		
+
 		if (pw->pw_gid == grp->gr_gid)
-                        return rush_true;
+			return rush_true;
 
 		for (p = grp->gr_mem; *p; p++) {
 			if (strcmp(*p, pw->pw_name) == 0)
 				return rush_true;
 		}
-        }
-        return rush_false;
+	}
+	return rush_false;
 }
 
 static rush_bool_t
@@ -370,12 +370,108 @@ eval_member(struct test_node *node,
 	size_t i;
 
 	for (i = 0; node->v.groups[i]; i++) {
-		if (groupmember(node->v.groups[i], req->pw))
+		if (groupmember(getgrnam(node->v.groups[i]), req->pw))
 			return rush_true;
 	}
 	return rush_false;
 }
 
+static rush_bool_t
+rush_access(struct stat *st, struct rush_request *req, int bit)
+{
+	if (st->st_mode & bit)
+		return rush_true;
+
+	bit <<= 3;
+	if ((st->st_mode & bit) && groupmember(getgrgid(st->st_gid), req->pw))
+		return rush_true;
+
+	bit <<= 3;
+	if ((st->st_mode & bit) && st->st_uid == req->pw->pw_uid)
+		return rush_true;
+
+	return rush_false;
+}
+
+static rush_bool_t
+eval_fstest(struct test_node *node,
+	    struct rush_rule *rule, struct rush_request *req)
+{
+	char *str = rush_expand_string(node->v.fstest.arg, req);
+	struct stat st;
+
+	if (((node->v.fstest.op == fs_symlink) ? lstat : stat)(str, &st)) {
+		if (errno == ENOENT) {
+			free(str);
+			return rush_false;
+		}
+		die(system_error, &req->i18n,
+		    _("%s:%d: %s: can't stat '%s': %s"),
+		    rule->file, rule->line, rule->tag,
+		    str,
+		    strerror(errno));
+	}
+	free(str);
+
+	switch (node->v.fstest.op) {
+	case fs_block_special:
+		return S_ISBLK(st.st_mode);
+
+	case fs_char_special:
+		return S_ISCHR(st.st_mode);
+
+	case fs_directory:
+		return S_ISDIR(st.st_mode);
+
+	case fs_exists:
+		return rush_true;
+
+	case fs_regular:
+		return S_ISREG(st.st_mode);
+
+	case fs_set_gid:
+		return st.st_mode & S_ISGID;
+
+	case fs_owner_egid:
+		return st.st_gid == req->pw->pw_gid;
+
+	case fs_symlink:
+		return S_ISLNK(st.st_mode);
+
+	case fs_sticky:
+		return st.st_mode & S_ISVTX;
+
+	case fs_owner_euid:
+		return st.st_uid == req->pw->pw_uid;
+
+	case fs_pipe:
+		return S_ISFIFO(st.st_mode);
+
+	case fs_readable:
+		return rush_access(&st, req, 4);
+
+	case fs_size:
+		return st.st_size > 0;
+
+	case fs_socket:
+		return S_ISSOCK(st.st_mode);
+
+	case fs_set_uid:
+		return st.st_mode & S_ISUID;
+
+	case fs_writable:
+		return rush_access(&st, req, 2);
+
+	case fs_executable:
+		return rush_access(&st, req, 1);
+
+	default:
+		die(system_error, NULL,
+		    _("INTERNAL ERROR at %s:%d: unrecognized operation %d"),
+		    __FILE__, __LINE__, node->v.fstest.op);
+	}
+}
+
 rush_bool_t
 test_eval(struct test_node *node,
 	  struct rush_rule *rule, struct rush_request *req)
@@ -383,7 +479,7 @@ test_eval(struct test_node *node,
 	switch (node->type) {
 	case test_cmpn:
 		return eval_cmpn(node, rule, req);
-		
+
 	case test_cmps:
 		return eval_cmps(node, rule, req);
 
@@ -392,12 +488,12 @@ test_eval(struct test_node *node,
 
 	case test_group:
 		return eval_member(node, rule, req);
-		
+
 	case test_and:
 		if (!test_eval(node->v.arg[0], rule, req))
 			return 0;
 		return test_eval(node->v.arg[1], rule, req);
-		
+
 	case test_or:
 		if (test_eval(node->v.arg[0], rule, req))
 			return 1;
@@ -406,6 +502,9 @@ test_eval(struct test_node *node,
 	case test_not:
 		return ! test_eval(node->v.arg[0], rule, req);
 
+	case test_fstest:
+		return eval_fstest(node, rule, req);
+
 	default:
 		die(system_error, NULL,
 		    _("INTERNAL ERROR at %s:%d: unrecognized node type %d"),
@@ -425,25 +524,25 @@ make_file_name(const char *dir, const char *name)
 	strcpy(res + dlen, name);
 	return res;
 }
- 
+
 char *
 expand_tilde(const char *dir, const char *home)
 {
-        char *res;
-        if (dir[0] == '~') {
-                if (dir[1] == '/') {
-                        size_t hlen = strlen(home);
-                        size_t len = hlen + strlen(dir + 1);
-                        res = xmalloc(len + 1);
-                        strcpy(res, home);
-                        if (hlen > 0 && res[hlen-1] != '/')
-                                res[hlen++] = '/';
-                        strcpy(res + hlen, dir + 2);
-                } else
-                        res = xstrdup(home);
-        } else
-                res = xstrdup(dir);
-        return res;
+	char *res;
+	if (dir[0] == '~') {
+		if (dir[1] == '/') {
+			size_t hlen = strlen(home);
+			size_t len = hlen + strlen(dir + 1);
+			res = xmalloc(len + 1);
+			strcpy(res, home);
+			if (hlen > 0 && res[hlen-1] != '/')
+				res[hlen++] = '/';
+			strcpy(res + hlen, dir + 2);
+		} else
+			res = xstrdup(home);
+	} else
+		res = xstrdup(dir);
+	return res;
 }
 
 char *
@@ -499,7 +598,7 @@ getenvind(struct rush_request *req, char const *name, char **pval)
      EV    - a pointer to an envar entry,
      VAR   - an entry from the environ array,
      LEN   - length of the variable name part of VAR (in other words,
-             position of the first = character in VAR).
+	     position of the first = character in VAR).
 */
 static enum rush_three_state
 envarmatch(struct envar *ev, char const *var, int len)
@@ -510,7 +609,7 @@ envarmatch(struct envar *ev, char const *var, int len)
 				? rush_true
 				: rush_false;
 		}
-	} else if (wildmatch(ev->name, var, len) == 0) {
+	} else if (wordsplit_wildmatch(ev->name, var, len) == 0) {
 		return rush_true;
 	}
 	return rush_undefined;
@@ -561,7 +660,7 @@ env_setup(struct rush_rule *rule, struct rush_request *req)
 {
 	struct envar *ev;
 	size_t i;
-	
+
 	if (rule->clrenv) {
 		size_t keep_count = 0;
 		for (i = 0; i < req->env_count; i++) {
@@ -583,7 +682,7 @@ env_setup(struct rush_rule *rule, struct rush_request *req)
 		char *val;
 		ssize_t n;
 		size_t len;
-		
+
 		switch (ev->type) {
 		case envar_keep:
 			/* Skip it */
@@ -614,30 +713,30 @@ env_setup(struct rush_rule *rule, struct rush_request *req)
 		case envar_eval:
 			free(rush_expand_string(ev->value, req));
 			break;
-			
+
 		default:
 			die(system_error, NULL,
 			    _("INTERNAL ERROR at %s:%d: invalid envar type %d"),
 			    __FILE__, __LINE__, ev->type);
 		}
 	}
-        if (__debug_p(2)) {
-                logmsg(LOG_DEBUG, _("Final environment:"));
-                for (i = 0; req->env[i]; i++)
-                        logmsg(LOG_DEBUG, "%4zu: %s", i, req->env[i]);
-        }
+	if (__debug_p(2)) {
+		logmsg(LOG_DEBUG, _("Final environment:"));
+		for (i = 0; req->env[i]; i++)
+			logmsg(LOG_DEBUG, "%4zu: %s", i, req->env[i]);
+	}
 }
 
 static void
 reparse_cmdline(struct rush_request *req)
 {
-        struct wordsplit ws;
-	
-        argcv_free(req->argc, req->argv);
+	struct wordsplit ws;
+
+	argcv_free(req->argc, req->argv);
 	ws.ws_options = WRDSO_NOVARSPLIT | WRDSO_NOCMDSPLIT;
 	if (wordsplit(req->cmdline, &ws, WRDSF_DEFFLAGS|WRDSF_OPTIONS))
 		die(system_error, &req->i18n, _("wordsplit(%s) failed: %s"),
-                    req->cmdline, wordsplit_strerror(&ws));
+		    req->cmdline, wordsplit_strerror(&ws));
 	wordsplit_get_words(&ws, &req->argc, &req->argv);
 	wordsplit_free(&ws);
 
@@ -648,8 +747,8 @@ reparse_cmdline(struct rush_request *req)
 static void
 rebuild_cmdline(struct rush_request *req)
 {
-        free(req->cmdline);
-        req->cmdline = argcv_string(req->argc, req->argv);
+	free(req->cmdline);
+	req->cmdline = argcv_string(req->argc, req->argv);
 }
 
 
@@ -661,7 +760,7 @@ get_arg_no(int index, struct rush_request *req)
 	int arg_no = ARG_NO(index, req->argc);
 	if (arg_no < 0 || arg_no > req->argc)
 		die(config_error,
-		    &req->i18n, 
+		    &req->i18n,
 		    _("no argument at index %d in command: %s"),
 		    index,
 		    req->cmdline);
@@ -776,13 +875,16 @@ remove_option(struct rush_request *req, struct option_defn *opt)
 					   option cluster. */
 					continue;
 				}
+				/* Advance to the character following the
+				   option name */
+				p++;
 			} else {
 				/* Skip unrecognized short option */
 				continue;
 			}
 
 			/* Remove the option (and its parameter, if any) */
-			if (remove_optarg(req, opt, i, p + 1)) {
+			if (remove_optarg(req, opt, i, p)) {
 				i--;
 				mod = 1;
 			}
@@ -800,7 +902,7 @@ rush_transform(struct transform_node *node, struct rush_request *req)
 	char *newval = NULL;
 	int arg_no;
 	void (*postprocess)(struct rush_request *) = NULL;
-	
+
 	if (node->type == transform_remopt) {
 		debug(2, _("Removing option %s %s"), node->v.remopt.s_opt,
 		      node->v.remopt.l_opt ? node->v.remopt.l_opt : "(null)");
@@ -816,7 +918,7 @@ rush_transform(struct transform_node *node, struct rush_request *req)
 		postprocess = reparse_cmdline;
 		debug(2, "%s", _("Transforming command line"));
 		break;
-		
+
 	case target_program:
 		/* Executable program name */
 		target_ptr = &req->prog;
@@ -849,7 +951,7 @@ rush_transform(struct transform_node *node, struct rush_request *req)
 		debug(2, _("Transforming variable %s=%s"),
 		      node->target.v.name, target_src ? target_src : "(null)");
 		break;
-		
+
 	case target_env:
 		/* Environment variable */
 		die(system_error, NULL,
@@ -894,9 +996,10 @@ rush_transform(struct transform_node *node, struct rush_request *req)
 		    __FILE__, __LINE__, node->type);
 		break;
 	}
-	
+
 	free(*target_ptr);
 	*target_ptr = newval;
+	debug(2, _("transformed value: %s"), newval);
 
 	if (postprocess)
 		postprocess(req);
@@ -906,7 +1009,7 @@ static void
 rush_transform_delete(struct transform_node *node, struct rush_request *req)
 {
 	int arg_no, arg_end, i;
-	
+
 	switch (node->target.type) {
 	case target_arg:
 		/* Single command line argument */
@@ -921,7 +1024,7 @@ rush_transform_delete(struct transform_node *node, struct rush_request *req)
 		if (arg_no == 0 || arg_end == 0)
 			die(config_error,
 			    &req->i18n, _("Deleting argv[0] is prohibited"));
-		for (i = arg_no; i <= arg_end; i++) 
+		for (i = arg_no; i <= arg_end; i++)
 			free(req->argv[i]);
 		i = arg_end - arg_no + 1;
 		memmove(req->argv + arg_no,
@@ -934,7 +1037,7 @@ rush_transform_delete(struct transform_node *node, struct rush_request *req)
 	case target_var:
 		rush_request_delvar(req, node->target.v.name);
 		break;
-		
+
 	case target_env:
 		/* Environment variable */
 		die(system_error, NULL,
@@ -950,21 +1053,21 @@ rush_transform_delete(struct transform_node *node, struct rush_request *req)
 static void
 run_transforms(struct rush_rule *rule, struct rush_request *req)
 {
-        struct transform_node *node;
-	
-        for (node = rule->transform_head; node; node = node->next) {
+	struct transform_node *node;
+
+	for (node = rule->transform_head; node; node = node->next) {
 		if (node->type == transform_delete)
 			rush_transform_delete(node, req);
 		else
 			rush_transform(node, req);
-	}			
-        if (__debug_p(2)) {
-                int i;
+	}
+	if (__debug_p(2)) {
+		int i;
 		logmsg(LOG_DEBUG, _("Program name: %s"), PROGFILE(req));
-                logmsg(LOG_DEBUG, _("Final arguments:"));
-                for (i = 0; i < req->argc; i++)
-                        logmsg(LOG_DEBUG, "% 4d: %s", i, req->argv[i]);
-        }
+		logmsg(LOG_DEBUG, _("Final arguments:"));
+		for (i = 0; i < req->argc; i++)
+			logmsg(LOG_DEBUG, "% 4d: %s", i, req->argv[i]);
+	}
 }
 
 static void
@@ -978,7 +1081,7 @@ acct_on(struct rush_rule *rule, struct rush_request *req, pid_t pid)
 	wtmp.command = req->cmdline;
 	if (rushdb_start(&wtmp))
 		die(system_error,
-		    &req->i18n, 
+		    &req->i18n,
 		    _("error writing to database %s: %s"),
 		    RUSH_DB, strerror(errno));
 }
@@ -987,7 +1090,7 @@ static void
 acct_off(void)
 {
 	if (rushdb_stop())
-		logmsg(LOG_ERR, 
+		logmsg(LOG_ERR,
 		       _("error writing stop to database file %s: %s"),
 		       RUSH_DB, strerror(errno));
 	rushdb_close();
@@ -1000,15 +1103,15 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 	pid_t pid;
 
 	signal(SIGCHLD, SIG_DFL);
-	
+
 	pid = fork();
 
 	if (pid == 0) {
 		return;
 	}
-	
-	if (pid == -1) 
-		die(system_error, &req->i18n, 
+
+	if (pid == -1)
+		die(system_error, &req->i18n,
 		    _("%s:%d: %s: cannot fork: %s"),
 		    rule->file, rule->line, rule->tag,
 		    strerror(errno));
@@ -1030,13 +1133,14 @@ fork_process(struct rush_rule *rule, struct rush_request *req)
 		       rule->tag, WTERMSIG(status));
 	} else
 		logmsg(LOG_NOTICE, _("%s: subprocess terminated"), rule->tag);
-	if (req->acct == rush_true) 
+	if (req->acct == rush_true)
 		acct_off();
 	if (req->post_sockaddr)
 		post_socket_send(req->post_sockaddr, rule, req);
 	exit(0);
 }
 
+#if !HAVE_GETGROUPLIST
 static int
 membergid(gid_t gid, size_t gc, gid_t *gv)
 {
@@ -1046,18 +1150,44 @@ membergid(gid_t gid, size_t gc, gid_t *gv)
 			return 1;
 	return 0;
 }
+#endif
 
 static void
 get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
 {
 	size_t gidc = 0, n = 0;
 	gid_t *gidv = NULL;
-	struct group *gr;
-	
+	gid_t group = req->gid == NO_GID ? req->pw->pw_gid : req->gid;
+
 	n = 32;
 	gidv = xcalloc(n, sizeof(gidv[0]));
 
-	gidv[0] = req->gid == NO_GID ? req->pw->pw_gid : req->gid;
+#if HAVE_GETGROUPLIST
+	int max_groups = (int) n;
+	for (;;) {
+		int ng = getgrouplist(req->pw->pw_name, group,
+				      gidv, &max_groups);
+		if (ng >= 0) {
+			/* On some systems getgrouplist returns 0 on success.
+			 */
+			gidc = ng > 0 ? ng : max_groups;
+			break;
+		}
+
+		/* Work around a bug on Darwin, where getgrouplist
+		   fails to increase max_groups. */
+		if (n == max_groups) {
+			if (INT_MAX / 3 * 2 <= n)
+				xalloc_die ();
+			max_groups += (max_groups + 1) / 2;
+		}
+		n = max_groups;
+		gidv = xrealloc(gidv, n * sizeof (gidv[0]));
+	}
+#else
+	struct group *gr;
+
+	gidv[0] = group;
 	gidc = 1;
 
 	setgrent();
@@ -1075,6 +1205,7 @@ get_user_groups(struct rush_request *req, size_t *pgidc, gid_t **pgidv)
 			}
 	}
 	endgrent();
+#endif
 	*pgidc = gidc;
 	*pgidv = gidv;
 }
@@ -1084,21 +1215,21 @@ setowner(struct rush_request *req)
 {
 	size_t gidc;
 	gid_t *gidv;
-	
+
 	get_user_groups(req, &gidc, &gidv);
 	if (setgroups(gidc, gidv) < 0)
 		die(system_error, &req->i18n,
 		    "setgroups: %s", strerror(errno));
 	if (setgid(gidv[0]))
 		die(system_error, &req->i18n, _("cannot enforce gid %lu: %s"),
-                    (unsigned long) gidv[0], strerror(errno));
+		    (unsigned long) gidv[0], strerror(errno));
 	free(gidv);
 
-        if (setuid(req->pw->pw_uid))
-                die(system_error, &req->i18n, _("cannot enforce uid %lu: %s"),
-                    (unsigned long) req->pw->pw_uid, strerror(errno));
+	if (setuid(req->pw->pw_uid))
+		die(system_error, &req->i18n, _("cannot enforce uid %lu: %s"),
+		    (unsigned long) req->pw->pw_uid, strerror(errno));
 
-	if (req->pw->pw_uid && setuid(0) == 0) 
+	if (req->pw->pw_uid && setuid(0) == 0)
 		die(system_error, &req->i18n,
 		    _("seteuid(0) succeeded when it should not"));
 }
@@ -1106,11 +1237,11 @@ setowner(struct rush_request *req)
 static void
 run_rule(struct rush_rule *rule, struct rush_request *req)
 {
-        debug(2, _("Rule %s at %s:%d matched"),
+	debug(2, _("Rule %s at %s:%d matched"),
 	      rule->tag, rule->file, rule->line);
 
 	env_setup(rule, req);
-        
+
 	if (rule->i18n.text_domain)
 		req->i18n.text_domain = rule->i18n.text_domain;
 	if (rule->i18n.localedir)
@@ -1118,31 +1249,31 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 	if (rule->i18n.locale)
 		req->i18n.locale = rule->i18n.locale;
 
-        if (rule->error) {
+	if (rule->error) {
 		const char *msg = rush_error_msg(rule->error, &rule->i18n);
-                debug(2, _("Error message: %s"), msg);
-                if (write(rule->error->fd, msg, strlen(msg)) < 0)
-                        die(system_error, &req->i18n, 
-                            _("Error sending diagnostic message to descriptor %d: %s"),
-                            rule->error->fd, strerror(errno));
-                exit(1);
-        }
-                        
-        if (set_user_limits (req->pw->pw_name, rule->limits))
-                die(usage_error, &req->i18n, _("cannot set limits for %s"), 
-                    req->pw->pw_name);
+		debug(2, _("Error message: %s"), msg);
+		if (write(rule->error->fd, msg, strlen(msg)) < 0)
+			die(system_error, &req->i18n,
+			    _("Error sending diagnostic message to descriptor %d: %s"),
+			    rule->error->fd, strerror(errno));
+		exit(1);
+	}
 
-        run_transforms(rule, req);
+	if (set_user_limits (req->pw->pw_name, rule->limits))
+		die(usage_error, &req->i18n, _("cannot set limits for %s"),
+		    req->pw->pw_name);
 
-        if (rule->chroot_dir) {
-                char *dir = expand_dir(rule->chroot_dir, req);
-                debug(2, _("Chroot dir: %s"), dir);
+	run_transforms(rule, req);
+
+	if (rule->chroot_dir) {
+		char *dir = expand_dir(rule->chroot_dir, req);
+		debug(2, _("Chroot dir: %s"), dir);
 		free(req->chroot_dir);
 		req->chroot_dir = dir;
-        }
-        if (rule->home_dir) {
-                free(req->home_dir);
-                req->home_dir = expand_dir(rule->home_dir, req);
+	}
+	if (rule->home_dir) {
+		free(req->home_dir);
+		req->home_dir = expand_dir(rule->home_dir, req);
 		debug(2, _("Home dir: %s"), req->home_dir);
 	}
 
@@ -1150,37 +1281,37 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		req->gid = rule->gid;
 		debug(2, _("GID: %lu"), (unsigned long) req->gid);
 	}
-	
+
 	if (rule->post_sockaddr.len)
 		req->post_sockaddr = &rule->post_sockaddr;
-	
+
 	if (rule->acct != rush_undefined)
 		req->acct = rule->acct;
-	
+
 	if (req->acct == rush_true)
 		req->fork = rush_true;
 	else if (rule->post_sockaddr.len)
 		req->fork = rush_true;
-	else if (rule->fork != rush_undefined) 
+	else if (rule->fork != rush_undefined)
 		req->fork = rule->fork;
 
-	if (rule->mask != NO_UMASK) 
+	if (rule->mask != NO_UMASK)
 		req->umask = rule->mask;
 
-        if (rule->fall_through)
-                return;
+	if (rule->fall_through)
+		return;
 
 	if (req->acct == rush_true &&
-	    rushdb_open(RUSH_DB, 1) != rushdb_result_ok) 
-		die(system_error, &req->i18n, 
+	    rushdb_open(RUSH_DB, 1) != rushdb_result_ok)
+		die(system_error, &req->i18n,
 		    _("cannot open database %s: %s"),
 		    RUSH_DB, rushdb_error_string);
-	
+
 	if (req->chroot_dir) {
 		uid_t uid;
 		struct passwd *pw;
-		
-		if (chroot(req->chroot_dir)) 
+
+		if (chroot(req->chroot_dir))
 			die(system_error, &req->i18n,
 			    _("cannot chroot to %s: %s"),
 		    req->chroot_dir, strerror(errno));
@@ -1193,32 +1324,32 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		req->pw = pw;
 	}
 
-        if (req->home_dir) {
+	if (req->home_dir) {
 		debug(2, _("chdir %s"), req->home_dir);
-		if (chdir(req->home_dir)) 
+		if (chdir(req->home_dir))
 			die(system_error, &req->i18n,
 			    _("cannot change to dir %s: %s"),
-                    req->home_dir, strerror(errno));
+		    req->home_dir, strerror(errno));
 	}
 
-        debug(2, _("Executing %s, %s"), PROGFILE(req), req->cmdline);
+	debug(2, _("Executing %s, %s"), PROGFILE(req), req->cmdline);
 	if (lint_option) {
 		if (dump_option)
 			dump_request(req, stdout);
 		exit(0);
 	}
 
-	if (req->fork == rush_true) 
+	if (req->fork == rush_true)
 		fork_process(rule, req);
 
 	setowner(req);
-	
-        umask(req->umask);
 
-        execve(PROGFILE(req), req->argv, req->env);
-        die(system_error, &req->i18n, _("%s:%d: %s: cannot execute %s: %s"),
-            rule->file, rule->line, rule->tag,
-            req->cmdline, strerror(errno));
+	umask(req->umask);
+
+	execve(PROGFILE(req), req->argv, req->env);
+	die(system_error, &req->i18n, _("%s:%d: %s: cannot execute %s: %s"),
+	    rule->file, rule->line, rule->tag,
+	    req->cmdline, strerror(errno));
 }
 
 
@@ -1226,49 +1357,227 @@ static char *command = NULL;
 static char *test_user_name = NULL;
 static int interactive;
 
-#include "rushopt.h"
+enum {
+    OPT_SHOW_DEFAULT = 256,
+};
+
+static struct optdef options[] = {
+	{
+		.opt_name = "debug",
+		.opt_argdoc = N_("NUMBER"),
+		.opt_doc = N_("Set debugging level."),
+	},
+	{
+		.opt_name = "d",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "test",
+		.opt_doc = N_("Run in test mode."),
+	},
+	{
+		.opt_name = "t",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+	{
+		.opt_name = "lint",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "T",
+		.opt_doc = N_("Scanner test.")
+	},
+
+	{
+		.opt_name = "trace",
+		.opt_doc = N_("Print grammar and lexical analyzer traces"),
+	},
+	{
+		.opt_name = "x",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "user",
+		.opt_argdoc = N_("NAME"),
+		.opt_doc = N_("Supply user name in test mode."),
+	},
+	{
+		.opt_name = "u",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "c",
+		.opt_argdoc = N_("COMMAND"),
+		.opt_doc = N_("Execute COMMAND.")
+	},
+
+	{
+		.opt_name = "i",
+		.opt_doc = N_("Force interactive shell.")
+	},
+
+	{
+		.opt_name = "dump",
+		.opt_argdoc = N_("KEYWORDS"),
+		.opt_doc = N_("Dump final request in test mode."),
+	},
+	{
+		.opt_name = "D",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "security-check",
+		.opt_argdoc = N_("CHECK"),
+		.opt_doc = N_("Add or remove configuration security check."),
+	},
+	{
+		.opt_name = "C",
+		.opt_flags = OPTFLAG_ALIAS
+	},
+
+	{
+		.opt_name = "show-default",
+		.opt_doc = N_("Show default configuration."),
+		.opt_code = OPT_SHOW_DEFAULT
+	},
+
+	{ NULL }
+}, *optdef[] = { options, NULL };
+
+static void
+setopt(struct parseopt *po, int code, char *arg)
+{
+	switch (code) {
+	case 'd':
+		debug_level = atoi(arg);
+		debug_option = 1;
+		break;
+
+	case 't':
+		lint_option = 1;
+		break;
+
+	case 'T':
+		scanner_test = 1;
+		break;
+
+	case 'x':
+		parser_traces++;
+		break;
+
+	case 'u':
+		lint_option = 1;
+		if (getuid())
+			die(usage_error,
+			    NULL,
+			    _("the --user option is allowed "
+			      "for the superuser only"));
+		test_user_name = arg;
+		break;
+
+	case 'c':
+		command = arg;
+		break;
+
+	case 'i':
+		lint_option = 1;
+		interactive = 1;
+		break;
+
+	case 'D':
+		dump_option = arg;
+		lint_option = 1;
+		break;
+
+	case 'C':
+		if (cfck_keyword(arg))
+			die(usage_error, NULL, _("unknown keyword: %s"), arg);
+		break;
+
+	case OPT_SHOW_DEFAULT:
+#ifdef RUSH_DEFAULT_CONFIG
+		printf("%s\n", RUSH_DEFAULT_CONFIG);
+		exit(0);
+#else
+		po->po_error(po, PO_MSG_ERR, "%s",
+			     _("No default configuration"));
+		exit(1);
+#endif
+		break;
+
+	default:
+		po->po_error(po, PO_MSG_ERR,
+			     "internal error at %s:%d\n", __FILE__, __LINE__);
+		abort();
+	}
+}
+
+extern void version_hook(WORDWRAP_FILE wf, struct parseopt *po);
+
+static struct parseopt po = {
+	.po_descr = N_("rush - restricted user shell."),
+	.po_argdoc = N_("[FILE]"),
+	.po_optdef = optdef,
+	.po_setopt = setopt,
+	.po_ex_usage = 1,
+	.po_package_name = PACKAGE_NAME,
+	.po_package_url = PACKAGE_URL,
+	.po_bugreport_address = PACKAGE_BUGREPORT,
+	.po_version_hook = version_hook
+};
 
 int
 main(int argc, char **argv)
 {
-        uid_t uid;
-        struct rush_rule *rule;
-        struct rush_request req;
+	uid_t uid;
+	int oargc = argc;
+	char **oargv = argv;
+	struct rush_rule *rule;
+	struct rush_request req;
 
 	rush_set_program_name(argv[0]);
 	rush_i18n_init();
-        umask(~(mode_t)0);
-        
-        openlog(program_name, LOG_NDELAY|LOG_PID, LOG_AUTHPRIV);
+	umask(~(mode_t)0);
+
+	openlog(program_name, LOG_NDELAY|LOG_PID, LOG_AUTHPRIV);
+
+	if (parseopt_getopt(&po, argc, argv) == OPT_ERR)
+		die(system_error, NULL, "%s", _("parseopt_getopt failed"));
+	parseopt_argv(&po, &argc, &argv);
+	parseopt_free(&po);
 
-	get_options(argc, argv);
 	cfgram_debug(parser_traces > 0);
 	cflex_debug(parser_traces > 1);
 
 	if (scanner_test) {
 		cfck_keyword("none");
-		if (argc > optind + 1) {
+		if (argc > 1) {
 			logmsg(LOG_ERR, "%s", _("too many arguments"));
 			exit(1);
 		}
-		cflex_test(argv[optind]);
+		cflex_test(argv[0]);
 		exit(0);
 	}
 
-	if (argc == optind + 1) {
+	if (argc == 1) {
 		if (lint_option)
-			rush_config_file = argv[optind];
+			rush_config_file = argv[0];
 		else
 			die(usage_error, NULL, _("invalid command line"));
-	} else if (argc > optind)
+	} else if (argc > 0)
 		die(usage_error, NULL, _("invalid command line"));
-	
+
 	/* Relinquish root privileges in test mode */
 	if (lint_option) {
 		if (setuid(getuid()))
 			die(system_error, NULL, "setuid: %s", strerror(errno));
 	}
-	
+
 	if (test_user_name) {
 		struct passwd *pw = getpwnam(test_user_name);
 		if (!pw)
@@ -1278,30 +1587,30 @@ main(int argc, char **argv)
 			    strerror(errno));
 	}
 
-        uid = getuid();
-        if ((rush_pw = getpwuid(uid)) == NULL)
-                die(system_error, NULL,
+	uid = getuid();
+	if ((rush_pw = getpwuid(uid)) == NULL)
+		die(system_error, NULL,
 		    _("invalid uid %lu"), (unsigned long) uid);
 
-        debug(2, _("user %s, uid %lu"), rush_pw->pw_name,
+	debug(2, _("user %s, uid %lu"), rush_pw->pw_name,
 	      (unsigned long) rush_pw->pw_uid);
 
-        cfparse();
+	cfparse();
 
 	if (!command) {
-		if (lint_option && !interactive) 
+		if (lint_option && !interactive)
 			exit(0);
 	}
 
-        if (__debug_p(2)) {
-                int i;
-                logmsg(LOG_DEBUG, _("Command line:"));
-                for (i = 0; i < argc; i++)
-                        logmsg(LOG_DEBUG, "% 4d: %s", i, argv[i]);
-                logmsg(LOG_DEBUG, _("Environment:"));
-                for (i = 0; environ[i]; i++)
-                        logmsg(LOG_DEBUG, "% 4d %s", i, environ[i]);
-        }
+	if (__debug_p(2)) {
+		int i;
+		logmsg(LOG_DEBUG, _("Command line:"));
+		for (i = 0; i < oargc; i++)
+			logmsg(LOG_DEBUG, "% 4d: %s", i, oargv[i]);
+		logmsg(LOG_DEBUG, _("Environment:"));
+		for (i = 0; environ[i]; i++)
+			logmsg(LOG_DEBUG, "% 4d %s", i, environ[i]);
+	}
 
 	memset(&req, 0, sizeof(req));
 	if (!command) {
@@ -1312,21 +1621,21 @@ main(int argc, char **argv)
 	req.cmdline = xstrdup(command);
 	request_set_env(&req);
 	reparse_cmdline(&req);
-	
-        req.pw = rush_pw;
+
+	req.pw = rush_pw;
 	req.umask = 022;
 	req.chroot_dir = NULL;
-        req.home_dir = NULL;
+	req.home_dir = NULL;
 	req.gid = NO_GID;
 	req.fork = rush_undefined;
 	req.acct = rush_undefined;
-	
+
 	for (rule = rule_head; rule; rule = rule->next) {
 		if (req.interactive != rule->interactive)
 			continue;
-                if (rule->test_node && !test_eval(rule->test_node, rule, &req))
+		if (rule->test_node && !test_eval(rule->test_node, rule, &req))
 			continue;
-                if (debug_level) {
+		if (debug_level) {
 			if (req.interactive)
 				logmsg(LOG_NOTICE,
 				       _("Serving interactive shell request for %s by rule %s"),
@@ -1336,10 +1645,10 @@ main(int argc, char **argv)
 				       _("Serving request \"%s\" for %s by rule %s"),
 				       command, req.pw->pw_name, rule->tag);
 		}
-                run_rule(rule, &req);
-        }
+		run_rule(rule, &req);
+	}
 	die(req.interactive ? nologin_error : usage_error, &req.i18n,
 	    _("no matching rule for \"%s\", user %s"),
 	    req.cmdline, req.pw->pw_name);
-        return 0;
+	return 0;
 }
